"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const LogManager_1 = require("../../log/LogManager");
const PromiseUtil_1 = require("../../util/PromiseUtil");
const IoCException_1 = require("../IoCException");
const Lifecycle_1 = require("../Lifecycle");
const SingletonDefinition_1 = require("./SingletonDefinition");
class BaseSingletonState extends Lifecycle_1.LifecycleState {
}
BaseSingletonState.INSTANTIATING = new BaseSingletonState('INSTANTIATING', 100);
BaseSingletonState.INSTANTIATED = new BaseSingletonState('INSTANTIATED', 200);
BaseSingletonState.PROPERTIES_SET = new BaseSingletonState('PROPERTIES_SET', 400);
exports.BaseSingletonState = BaseSingletonState;
var ParamType;
(function (ParamType) {
    ParamType[ParamType["Value"] = 0] = "Value";
    ParamType[ParamType["Reference"] = 1] = "Reference";
    ParamType[ParamType["Type"] = 2] = "Type";
    ParamType[ParamType["TypeArray"] = 3] = "TypeArray";
    ParamType[ParamType["Config"] = 4] = "Config";
})(ParamType = exports.ParamType || (exports.ParamType = {}));
class ParamDefinition {
    static withValue(val) {
        const resp = new ParamDefinition(ParamType.Value);
        resp.val = val;
        return resp;
    }
    static withConfigKey(key) {
        const resp = new ParamDefinition(ParamType.Config);
        resp.key = key;
        return resp;
    }
    static withRefName(refName) {
        const resp = new ParamDefinition(ParamType.Reference);
        resp.refName = refName;
        return resp;
    }
    static withClassName(className) {
        const resp = new ParamDefinition(ParamType.Type);
        resp.className = className;
        return resp;
    }
    static withClassNameArr(className) {
        const resp = new ParamDefinition(ParamType.TypeArray);
        resp.className = className;
        return resp;
    }
    constructor(type) {
        this.type = type;
    }
}
class CallDefinition {
    constructor(paramName, paramDefinition) {
        this.args = [];
        this.paramName = paramName;
        this.args.push(paramDefinition);
    }
}
class BaseSingletonDefinition extends SingletonDefinition_1.SingletonDefinition {
    constructor(clazz, name, logger) {
        super(clazz, name, logger || LogManager_1.LogManager.getLogger());
        this.constructorArgDefinitions = [];
        this.propertiesToSetDefinitions = [];
        this.propertiesToSetDefinitions = [];
        this.startFunctionName = null;
        this.shutdownFunctionName = null;
    }
    static getMaxStateInstance(def, trace, postLoad) {
        // console.log(`In ${this.getName()} getting max instance of ${def.getName()} - ${def.status} - ${BaseSingletonState.STARTED} / trace: [${trace}]`);
        if ((trace.indexOf(def.getName()) >= 0) || (def.status < BaseSingletonState.STARTED)) {
            // Circular reference, let's try to go for one that is just instantiated and finalise init afterwards
            postLoad.add(def);
            // console.log(`Getting ${def.getName()} as Instantiated`);
            return def.getInstanceAtState(BaseSingletonState.INSTANTIATED, trace.concat([def.getName()]), postLoad);
        }
        // console.log(`Getting ${def.getName()} as ready`);
        return def.getInstanceWithTrace(trace.concat([def.getName()]), postLoad);
    }
    // ************************************
    // Configuration methods
    // ************************************
    constructorParamByValue(value) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.constructorArgDefinitions.push(ParamDefinition.withValue(value));
        return this;
    }
    constructorParamByConfig(key) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.constructorArgDefinitions.push(ParamDefinition.withConfigKey(key));
        return this;
    }
    constructorParamByRef(name) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.constructorArgDefinitions.push(ParamDefinition.withRefName(name));
        return this;
    }
    constructorParamByType(clazz) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.constructorArgDefinitions.push(ParamDefinition.withClassName(clazz));
        return this;
    }
    constructorParamByTypeArray(clazz) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.constructorArgDefinitions.push(ParamDefinition.withClassNameArr(clazz));
        return this;
    }
    startFunction(startFunctionName) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        if (Object.prototype.hasOwnProperty.call(this.clazz.prototype, startFunctionName)) {
            this.startFunctionName = startFunctionName;
        }
        else {
            throw new IoCException_1.IoCException(`Can't find function named ${startFunctionName} on class ${this.clazz.name}`);
        }
        return this;
    }
    stopFunction(shutdownFunctionName) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        if (Object.prototype.hasOwnProperty.call(this.clazz.prototype, shutdownFunctionName)) {
            this.shutdownFunctionName = shutdownFunctionName;
        }
        else {
            throw new IoCException_1.IoCException(`Can't find function named ${shutdownFunctionName} on class ${this.clazz.name}`);
        }
        return this;
    }
    setPropertyByValue(paramName, value) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.propertiesToSetDefinitions.push(new CallDefinition(paramName, ParamDefinition.withValue(value)));
        return this;
    }
    setPropertyByConfig(paramName, key) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.propertiesToSetDefinitions.push(new CallDefinition(paramName, ParamDefinition.withConfigKey(key)));
        return this;
    }
    setPropertyByRef(paramName, name) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.propertiesToSetDefinitions.push(new CallDefinition(paramName, ParamDefinition.withRefName(name)));
        return this;
    }
    setPropertyByType(paramName, className) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.propertiesToSetDefinitions.push(new CallDefinition(paramName, ParamDefinition.withClassName((typeof className === 'function' && className.name) ?
            className.name : className)));
        return this;
    }
    setPropertyByTypeArray(paramName, className) {
        this.assertState(BaseSingletonState.NOT_STARTED);
        this.propertiesToSetDefinitions.push(new CallDefinition(paramName, ParamDefinition.withClassNameArr((typeof className === 'function' && className.name) ?
            className.name : className)));
        return this;
    }
    // ************************************
    // Get instance methods
    // ************************************
    getInstance() {
        // console.log(`Getting instance of ${this.getName()}`);
        const postLoad = new Set();
        return this.getInstanceWithTrace([this.getName()], postLoad)
            .then((instance) => PromiseUtil_1.PromiseUtil.map(Array.from(postLoad.values()), (d) => d.getInstance())
            .then(() => instance));
    }
    getInstanceWithTrace(trace, postLoad) {
        return this.getInstanceAtState(BaseSingletonState.STARTED, trace, postLoad);
    }
    getInstanceAtState(minState, trace, postLoad) {
        // console.log(`Getting instance of ${this.getName()} at state ${BaseSingletonState.fromValue(minState)}`);
        if (minState !== BaseSingletonState.INSTANTIATED && minState !== BaseSingletonState.STARTED) {
            return Promise.reject(new IoCException_1.IoCException(`Doesn't make sense to request an object in states other than INSTANTIATED and STARTED: ${minState}`));
        }
        if (this.status.isAtOrAfter(BaseSingletonState.STOPPING)) {
            return Promise.reject(new IoCException_1.IoCException(`Object ${this.getName()} is stopping. Can't give it to you`));
        }
        if (this.status === BaseSingletonState.INSTANTIATING) {
            return Promise.reject(new IoCException_1.IoCException(`Circular dependency detected. Can't instantiate ${this.getName()}`));
        }
        if (minState.isAtOrBefore(this.status)) {
            return Promise.resolve(this.instance);
        }
        let basePromise;
        if (this.status.isBefore(BaseSingletonState.INSTANTIATING)) {
            basePromise = this.instantiate(trace, postLoad);
        }
        else {
            basePromise = Promise.resolve();
        }
        // The object is, at least, instantiated.
        if (minState === BaseSingletonState.INSTANTIATED) {
            return basePromise.then(() => this.instance);
        }
        // Then the minState is Started. Let's finish up then
        return basePromise
            .then(() => this.setAllProperties(trace, postLoad))
            .then(() => this.lcStart())
            .then(() => this.instance);
    }
    // ************************************
    // Lifecycle methods
    // ************************************
    doStart() {
        if (this.startFunctionName) {
            const resp = this.clazz.prototype[this.startFunctionName].call(this.instance);
            if (resp && resp.then) {
                return resp;
            }
            return Promise.resolve();
        }
        return Promise.resolve();
    }
    doStop() {
        if (this.shutdownFunctionName) {
            const resp = this.clazz.prototype[this.shutdownFunctionName].call(this.instance);
            if (resp && resp.then) {
                return resp;
            }
            return Promise.resolve();
        }
        return Promise.resolve();
    }
    resolveArgs(constructorArgs, trace, postLoad) {
        return PromiseUtil_1.PromiseUtil.map(constructorArgs, (arg) => {
            switch (arg.type) {
                case ParamType.Value:
                    return arg.val;
                case ParamType.Config:
                    return this.context.getConfig(arg.key);
                case ParamType.Reference:
                    return BaseSingletonDefinition.getMaxStateInstance(this.context.getDefinitionByName(arg.refName), trace, postLoad);
                case ParamType.Type:
                    return BaseSingletonDefinition.getMaxStateInstance(this.context.getDefinitionByType(arg.className), trace, postLoad);
                case ParamType.TypeArray:
                    return Promise.all(this.context.getDefinitionsByType(arg.className).map((d) => BaseSingletonDefinition.getMaxStateInstance(d, trace, postLoad)));
                default:
                    return Promise.reject(new IoCException_1.IoCException(`Unknown argument type ${arg.type} on bean ${this.name}`));
            }
        });
    }
    getCopyInstance() {
        return new BaseSingletonDefinition(this.clazz, this.name, this.logger);
    }
    copyInternalProperties(copyTo) {
        super.copyInternalProperties(copyTo);
        copyTo.constructorArgDefinitions = this.constructorArgDefinitions.slice(0);
        copyTo.propertiesToSetDefinitions = this.propertiesToSetDefinitions.slice(0);
        copyTo.startFunctionName = this.startFunctionName;
        copyTo.shutdownFunctionName = this.shutdownFunctionName;
    }
    instantiate(trace, postLoad) {
        if (!this.context) {
            return Promise.reject(new IoCException_1.IoCException(`ObjectDefinition ${this.getName()} hasn't been added to a context, Can't instantiate`));
        }
        this.setStatus(BaseSingletonState.INSTANTIATING);
        return this.resolveArgs(this.constructorArgDefinitions, trace, postLoad)
            .then((constructorArgs) => {
            this.instance = Reflect.construct(this.clazz, constructorArgs);
            this.setStatus(BaseSingletonState.INSTANTIATED);
            return this.instance;
        });
    }
    setAllProperties(trace, postload) {
        return PromiseUtil_1.PromiseUtil.map(this.propertiesToSetDefinitions, (propertyToSet) => this.resolveArgs(propertyToSet.args, trace, postload).then((args) => {
            this.instance[propertyToSet.paramName] = args[0];
        }));
    }
}
exports.BaseSingletonDefinition = BaseSingletonDefinition;
//# sourceMappingURL=BaseSingletonDefinition.js.map